import arrow

try:
    PANDAS_IS_INSTALLED = True
    import pandas as pd
except ImportError:
    PANDAS_IS_INSTALLED = False


def timestamp_converter(ts):
    if PANDAS_IS_INSTALLED:
        if isinstance(ts, arrow.Arrow):
            ts = ts.isoformat()
        ts = pd.Timestamp(ts)
        if not ts.tz:
            ts = ts.tz_localize('UTC')
        return ts
    else:
        return arrow.get(ts)


def to_dataframe(self):
    return pd.DataFrame(data={'value': self.values()}, index=self.keys())



import os
from pathlib import Path

import sphinx_rtd_theme

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

# Embed Ipython configuration:
ipython_mplbackend = ""

# Execute ipython_startup_file before each ipython embed
ipython_startup_file = Path(__file__).parent / "ipython_init.py"
ipython_execlines = ["%run {}".format(ipython_startup_file.as_posix())]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['.']
templates_path = ["_templates"]

if os.getenv("SPELLCHECK"):
    extensions += ("sphinxcontrib.spelling",)
    spelling_show_suggestions = True
    spelling_lang = "en_US"

# The suffix(es) of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"ticts"
author = "GEM"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Disable docstring inheritance
autodoc_inherit_docstrings = False

extlinks = {
    # 'issue': ('https://github.com/gjeusel/ticts/issues
    # 'pr': ('https://github.com/gjeusel/ticts/pulls
}


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "collapse_navigation": False,
    "display_version": True,
    "logo_only": True,
}

html_logo = "_static/img/logo.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_context = {
    "css_files": [
        "https://fonts.googleapis.com/css?family=Lato",
        "_static/css/custom_theme.css",
    ]
}



from datetime import timedelta

import pytest

from ticts import TimeSeries
from ticts.pandas_ext import timestamp_converter

CURRENT = timestamp_converter('2019-01-01')
ONEHOUR = timedelta(hours=1)
HALFHOUR = timedelta(minutes=30)
ONEMIN = timedelta(minutes=1)


@pytest.fixture
def smalldict():
    dct = dict()
    for i in range(10):
        dct[CURRENT + i * ONEHOUR] = i
    return dct


@pytest.fixture
def smallts(smalldict):
    return TimeSeries(smalldict)


@pytest.fixture
def smallts_withdefault(smalldict):
    return TimeSeries(smalldict, default=10)


@pytest.fixture
def otherict():
    return {
        CURRENT + 2 * ONEHOUR: 1000,
        CURRENT + 2 * ONEHOUR + 30 * ONEMIN: 2000,
        CURRENT + 4 * ONEHOUR: 3000,
    }


@pytest.fixture
def conditionalts():
    dct = {
        CURRENT + 2 * ONEHOUR: True,
        CURRENT + 3 * ONEHOUR: False,
    }
    return TimeSeries(dct, default=True)


@pytest.fixture
def otherts(otherict):
    return TimeSeries(otherict)


@pytest.fixture
def otherts_withdefault(otherict):
    return TimeSeries(otherict, default=900)


@pytest.fixture
def emptyts():
    return TimeSeries()


@pytest.fixture
def emptyts_withdefault():
    return TimeSeries(default=10)



import datetime
import logging
from pathlib import Path

import pytz

try:
    import pandas as pd

    pd.options.display.max_rows = 10

    emptydf = pd.DataFrame()

    # Recent Dates aware:
    now = pd.Timestamp.now(tz="CET").replace(microsecond=0)
    end = pd.Timestamp(now, tz="CET") + pd.Timedelta(hours=4)
    start = end - pd.Timedelta(days=3)

    # DST
    dst_start = pd.Timestamp("2017-03-26T00:00:00", tz="CET")
    dst_end = pd.Timestamp("2017-03-26T04:00:00", tz="CET")

    # 2016
    start_2016 = pd.Timestamp("2016-01-01T00:00:00", tz="CET")
    end_2016 = pd.Timestamp("2017-01-01T00:00:00", tz="CET")

    # 2017
    start_2017 = pd.Timestamp("2017-01-01T00:00:00", tz="CET")
    end_2017 = pd.Timestamp("2017-01-01T00:00:00", tz="CET")

    # 2018
    start_2018 = pd.Timestamp("2018-01-01T00:00:00", tz="CET")
    end_2018 = pd.Timestamp("2018-01-01T00:00:00", tz="CET")

    # timedeltas aliases:
    onehour = pd.Timedelta(hours=1)
    tenhours = pd.Timedelta(hours=10)
    oneday = pd.Timedelta(days=1)
    oneweek = pd.Timedelta(days=7)
    onemonth = pd.Timedelta(days=31)
except Exception as e:
    print("Could not import pandas as pd: {}".format(e))



from copy import copy, deepcopy
from datetime import timedelta
from unittest import mock

import pytest

from ticts import TimeSeries

from .conftest import CURRENT, HALFHOUR, ONEHOUR, ONEMIN


class TestTimeSeriesInit:
    def test_with_dict(self, smalldict):
        ts = TimeSeries(smalldict)
        assert ts[CURRENT + ONEHOUR] == 1

    def test_with_data_as_tuple(self):
        mytuple = ((CURRENT, 0), (CURRENT + ONEHOUR, 1))
        ts = TimeSeries(mytuple)
        assert ts[CURRENT] == 0
        assert ts[CURRENT + ONEHOUR] == 1
        assert len(ts) == 2


class TestTimeSeriesSetItem:
    def test_simple_setitem(self, smallts):
        smallts[CURRENT] = 1000
        assert smallts[CURRENT] == 1000

    def test_consecutive_setitem(self, smallts):
        smallts[CURRENT] = 1000
        first_time = deepcopy(smallts)
        smallts[CURRENT] = 1000
        assert first_time == smallts

    @mock.patch("ticts.TimeSeries.set_interval")
    def test_setitem_on_slice_calls_set_interval(self, set_interval, smallts):
        smallts[CURRENT:CURRENT + 2 * ONEHOUR] = 1000
        assert set_interval.call_count == 1


class TestTimeSeriesCopy:
    def test_copy(self, smallts):
        copied = copy(smallts)
        assert copied == smallts

    def test_copy_with_default(self, smallts_withdefault):
        copied = copy(smallts_withdefault)
        assert copied == smallts_withdefault

    def test_deepcopy(self, smallts):
        deepcopied = deepcopy(smallts)
        assert deepcopied == smallts

    def test_deepcopy_with_default(self, smallts_withdefault):
        deepcopied = deepcopy(smallts_withdefault)
        assert deepcopied == smallts_withdefault


def test_timeseries_compact(smallts):
    smallts[CURRENT + ONEMIN] = 0
    assert (CURRENT + ONEMIN) not in smallts.compact().keys()


class TestTimeSeriesGetitem:
    available_interpolate = ['previous', 'linear']

    # tests on corner cases

    def test_getitem_out_of_left_bound_with_no_default_raises(self, smallts):
        with pytest.raises(KeyError) as err:
            smallts[CURRENT - ONEMIN]

        assert 'default attribute is not set' in str(err)

    @pytest.mark.parametrize('interpolate', available_interpolate)
    def test_getitem_out_of_left_bound_with_default_return_default(
            self, smallts_withdefault, interpolate):
        value = smallts_withdefault[CURRENT - ONEMIN, interpolate]
        assert value == smallts_withdefault.default

    @pytest.mark.parametrize('interpolate', available_interpolate)
    def test_getitem_on_empty_when_no_default_raises(self, emptyts,
                                                     interpolate):
        with pytest.raises(KeyError) as err:
            emptyts[CURRENT - ONEMIN, interpolate]

        assert str(
            "default attribute is not set and timeseries is empty") in str(err)

    # tests on '_get_previous'

    @mock.patch("ticts.TimeSeries._get_previous")
    def test_get_on_previous_is_default_interpolate(self, _get_previous,
                                                    smallts):
        smallts[CURRENT + ONEMIN]
        assert _get_previous.call_count == 1

    def test_get_on_previous(self, smallts):
        assert smallts[CURRENT + ONEMIN] == 0

    def test_get_on_previous_out_of_right_bound(self, smallts):
        assert smallts[CURRENT + 10 * ONEHOUR] == 9

    # tests on '_get_linear_interpolate'

    @pytest.mark.parametrize('time_idx, expected', [
        (CURRENT + HALFHOUR, 0.5),
        (CURRENT + ONEHOUR + HALFHOUR, 1.5),
        (CURRENT + 10 * ONEMIN, 0 + (1 - 0) * (10 * ONEMIN / ONEHOUR)),
    ])
    def test_get_linear_interpolate(self, smallts, time_idx, expected):
        assert smallts[time_idx, 'linear'] == expected

    def test_get_linear_interpolate_out_of_right_bound(self, smallts):
        assert smallts[CURRENT + 10 * ONEHOUR, 'linear'] == 9

    # test on 'slice'

    def test_get_on_slice_exclude_upper_bound_include_lower_bound(
            self, smallts):
        start = CURRENT
        end = CURRENT + 1 * ONEHOUR
        data = {start: 0, end: 1}
        expected_ts = TimeSeries(data, default=smallts.default)
        assert smallts[start:end + ONEHOUR] == expected_ts

    def test_get_on_slice_add_back_previous_value_if_start_not_in_keys(
            self, smallts):
        start = CURRENT + HALFHOUR
        end = CURRENT + ONEHOUR
        data = {start: 0, end: 1}
        expected_ts = TimeSeries(data, default=smallts.default)
        assert smallts[start:end + ONEHOUR] == expected_ts

    def test_get_on_slice_entirely_out_of_bounds_on_left_side(self, smallts):
        assert smallts[CURRENT - 2 * ONEHOUR:CURRENT - 1 * ONEHOUR].empty

    def test_get_on_slice_out_of_bounds_left_side(self, smallts):
        start = CURRENT - 2 * ONEHOUR
        end = CURRENT
        data = {CURRENT: 0}
        expected_ts = TimeSeries(data, default=smallts.default)
        assert smallts[start:end + 1 * ONEHOUR] == expected_ts

    def test_get_on_slice_entirely_out_of_bounds_on_right_side(self, smallts):
        assert smallts[CURRENT + 10 * ONEHOUR:CURRENT + 12 * ONEHOUR].empty

    def test_get_on_slicout_of_bounds_on_right_side(self, smallts):
        sliced_ts = smallts[CURRENT + 9 * ONEHOUR:CURRENT + 12 * ONEHOUR]
        expected_dct = {CURRENT + 9 * ONEHOUR: 9}
        expected_ts = TimeSeries(expected_dct, default=smallts.default)
        assert sliced_ts == expected_ts


class TestTimeSeriesSetInterval:
    def test_set_interval_raises_when_no_default(self, smallts):
        with pytest.raises(NotImplementedError):
            smallts.set_interval(CURRENT, CURRENT + ONEHOUR, 1000)

    def test_single_set_interval_end_on_last_key(self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT + ONEHOUR,
                                         CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT, CURRENT + ONEHOUR, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.keys()) == expected_keys
        assert smallts_withdefault[CURRENT + ONEHOUR] == 1000

    def test_single_set_interval_start_on_first_key(self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT, CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.keys()) == expected_keys
        assert smallts_withdefault[CURRENT] == 1000

    @pytest.mark.parametrize('start, end', [
        (CURRENT + ONEHOUR, CURRENT + 10 * ONEHOUR),
        (CURRENT + 4 * ONEHOUR, CURRENT + 11 * ONEHOUR),
    ])
    def test_single_set_interval_end_over_last_key_sets_to_last_value(
            self, smallts_withdefault, start, end):
        last_key = smallts_withdefault.keys()[-1]
        last_val = smallts_withdefault[last_key]

        smallts_withdefault.set_interval(start, end, 1000)
        keys_before_start = []
        for key in smallts_withdefault.keys():
            if key < start:
                keys_before_start.append(key)

        expected_keys = [*keys_before_start, start, end]
        assert list(smallts_withdefault.keys()) == expected_keys
        assert smallts_withdefault[end] == last_val

    def test_single_set_interval_when_start_higher_than_upper_bound(
            self, smallts_withdefault):
        start = CURRENT + 11 * ONEHOUR
        end = CURRENT + 13 * ONEHOUR
        smallts_withdefault.set_interval(start, end, 1000)
        assert smallts_withdefault[CURRENT + 10 * ONEHOUR] == 9
        assert smallts_withdefault[start] == 1000
        assert smallts_withdefault[end] == smallts_withdefault.default

    def test_single_set_interval_start_before_first_key(
            self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT - ONEHOUR,
                                         CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT - ONEHOUR, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.keys()) == expected_keys
        assert smallts_withdefault[CURRENT - 1 * ONEHOUR] == 1000

    def test_single_set_interval_on_bounds_not_being_keys(
            self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT + ONEMIN,
                                         CURRENT + 9 * ONEHOUR, 1000)
        expected_keys = [CURRENT, CURRENT + ONEMIN, CURRENT + 9 * ONEHOUR]
        assert list(smallts_withdefault.keys()) == expected_keys
        assert smallts_withdefault[CURRENT + ONEMIN] == 1000

    def test_set_interval_on_empty(self, emptyts):
        emptyts.default = 10
        emptyts.set_interval(CURRENT, CURRENT + ONEHOUR, 1)
        assert emptyts[CURRENT] == 1
        len(emptyts.keys()) == 1

    def test_set_interval_on_empty_with_default(self, emptyts_withdefault):
        emptyts_withdefault.set_interval(CURRENT, CURRENT + ONEHOUR, 1)
        assert emptyts_withdefault[CURRENT] == 1
        assert emptyts_withdefault[CURRENT +
                                   ONEHOUR] == emptyts_withdefault.default
        len(emptyts_withdefault.keys()) == 2

    def test_same_consecutive_set_interval(self, smallts_withdefault):
        smallts_withdefault.set_interval(CURRENT, CURRENT + 9 * ONEHOUR, 1000)
        first_time = deepcopy(smallts_withdefault)
        smallts_withdefault.set_interval(CURRENT, CURRENT + 9 * ONEHOUR, 1000)
        assert first_time == smallts_withdefault

    def test_consecutive_set_interval_on_empty_with_default(self, emptyts):
        emptyts.default = 10
        emptyts.set_interval(CURRENT, CURRENT + 2 * ONEHOUR, 10)
        emptyts.set_interval(CURRENT, CURRENT + ONEHOUR, 0)
        emptyts.set_interval(CURRENT + ONEHOUR, CURRENT + 2 * ONEHOUR, 1)
        emptyts.set_interval(CURRENT, CURRENT + 2 * ONEHOUR, 3)
        assert emptyts[CURRENT] == 3
        assert emptyts[CURRENT + 2 * ONEHOUR] == 10
        assert list(emptyts.keys()) == [CURRENT, CURRENT + 2 * ONEHOUR]

    def test_set_interval_when_no_keys_to_delete_with_default(self, emptyts):
        emptyts.default = 1000
        emptyts.set_interval(CURRENT, CURRENT + 1 * ONEHOUR, 0)
        emptyts.set_interval(CURRENT + 3 * ONEHOUR, CURRENT + 4 * ONEHOUR, 3)
        emptyts.set_interval(CURRENT + 1 * ONEHOUR + HALFHOUR,
                             CURRENT + 3 * ONEHOUR, 10)

        expected_dct = {
            CURRENT: 0,
            CURRENT + 1 * ONEHOUR: 1000.,
            CURRENT + 1 * ONEHOUR + HALFHOUR: 10,
            CURRENT + 3 * ONEHOUR: 3,
            CURRENT + 4 * ONEHOUR: 1000
        }
        assert list(emptyts.keys()) == list(expected_dct.keys())

        for key in expected_dct:
            assert emptyts[key] == expected_dct[key]


class TestTimeSeriesOperators:
    def test_simple_add(self, smallts, smalldict):
        ts = smallts + smallts
        newdct = {key: value + value for key, value in smalldict.items()}
        assert ts == TimeSeries(newdct)

    def test_simple_add_one_float(self, smallts, smalldict):
        ts = smallts + 1000
        assert list(ts.values()) == list(range(1000, 1010))

    def test_add_with_keys_differences(self, smallts_withdefault,
                                       otherts_withdefault):
        ts = smallts_withdefault + otherts_withdefault
        assert ts[CURRENT + 1 * ONEHOUR] == 1 + otherts_withdefault.default
        assert ts[CURRENT + 2 * ONEHOUR] == 2 + 1000
        assert ts[CURRENT + 2 * ONEHOUR + 30 * ONEMIN] == 2 + 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 + 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 + 3000

    def test_add_on_list_of_timeseries(self, smallts):
        lst_timeseries = [smallts, smallts, smallts]
        result = sum(lst_timeseries)

        for key in result:
            assert result[key] == 3 * smallts[key]

    def test_simple_sub(self, smallts):
        ts = smallts - smallts
        assert all([val == 0 for val in ts.values()])

    def test_simple_sub_one_float(self, smallts):
        ts = smallts - 1
        assert list(ts.values()) == list(range(-1, 9))

    def test_sub_with_keys_differences(self, smallts_withdefault,
                                       otherts_withdefault):
        ts = smallts_withdefault - otherts_withdefault
        assert ts[CURRENT + 1 * ONEHOUR] == 1 - 900
        assert ts[CURRENT + 2 * ONEHOUR] == 2 - 1000
        assert ts[CURRENT + 2 * ONEHOUR + 30 * ONEMIN] == 2 - 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 3 - 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 4 - 3000

    def test_simple_le(self, smallts):
        result = smallts >= 5
        assert all([
            not val for val in result[CURRENT:CURRENT + 4 * ONEHOUR].values()
        ])
        assert all([
            val for val in result[CURRENT + 5 * ONEHOUR:CURRENT +
                                  9 * ONEHOUR].values()
        ])

    def test_simple_lt(self, smallts):
        result = smallts > 5
        assert all([
            not val for val in result[CURRENT:CURRENT + 5 * ONEHOUR].values()
        ])
        assert all([
            val for val in result[CURRENT + 6 * ONEHOUR:CURRENT +
                                  9 * ONEHOUR].values()
        ])

    def test_floor_on_float(self, smallts):
        ts = smallts.floor(2)
        assert all([value <= 2 for value in ts.values()])

    def test_floor_on_ts(self, smallts_withdefault, otherts_withdefault):
        ts = smallts_withdefault.floor(otherts_withdefault)
        assert ts[CURRENT + 1 * ONEHOUR] == 1
        assert ts[CURRENT + 2 * ONEHOUR] == 2
        assert ts[CURRENT + 2 * ONEHOUR + 30 * ONEMIN] == 2
        assert ts[CURRENT + 3 * ONEHOUR] == 3
        assert ts[CURRENT + 4 * ONEHOUR] == 4

    def test_ceil_on_float(self, smallts):
        ts = smallts.ceil(7)
        assert all([value >= 7 for value in ts.values()])

    def test_ceil_on_ts(self, smallts_withdefault, otherts_withdefault):
        ts = smallts_withdefault.ceil(otherts_withdefault)
        assert ts[CURRENT + ONEHOUR] == 900
        assert ts[CURRENT + 2 * ONEHOUR] == 1000
        assert ts[CURRENT + 2 * ONEHOUR + 30 * ONEMIN] == 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 3000

    def test_ceil_on_ts_check_changing_default(self, smallts_withdefault,
                                               otherts_withdefault):
        otherts_withdefault.default = -10
        ts = smallts_withdefault.ceil(otherts_withdefault)

        assert ts[CURRENT + ONEHOUR] == 1
        assert ts[CURRENT + 2 * ONEHOUR] == 1000
        assert ts[CURRENT + 2 * ONEHOUR + 30 * ONEMIN] == 2000
        assert ts[CURRENT + 3 * ONEHOUR] == 2000
        assert ts[CURRENT + 4 * ONEHOUR] == 3000


class TestConditionalUpdate:
    def test_simple_conditional_update_raises_on_type_of_other(
            self, smallts, conditionalts):
        with pytest.raises(TypeError) as err:
            smallts.conditional_update("fakedct", conditionalts)

        assert "other should be of type TimeSeries, got" in str(err)

    def test_simple_conditional_update_raises_on_type_of_condition_ts(
            self, smallts):
        with pytest.raises(TypeError) as err:
            smallts.conditional_update(smallts, smallts)

        assert "The values of condition should all be boolean" in str(err)

    def test_simple_conditional_update(self, smallts, otherts, conditionalts):
        smallts.conditional_update(otherts, conditionalts)
        assert smallts[CURRENT + 2 * ONEHOUR] == 1000
        assert smallts[CURRENT + 3 * ONEHOUR] == 3
        assert smallts[CURRENT + 4 * ONEHOUR] == 4


class TestTimeSeriesSample:
    def test_simple_sample_on_default(self, smallts):
        ts = smallts.sample(HALFHOUR)
        expected_dict = {
            **smallts,
            **{
                key + HALFHOUR: smallts[key]
                for key in list(smallts.keys())[:-1]
            }
        }
        expected_ts = TimeSeries(expected_dict, default=smallts.default)
        assert ts == expected_ts

    def test_sample_raises_when_no_default_and_start_lower_than_all_keys(
            self, smallts):
        with pytest.raises(KeyError) as err:
            smallts.sample(freq=HALFHOUR, start=CURRENT - ONEHOUR)

        assert "default attribute is not set, can't deduce value" in str(err)

    def test_sample_with_start_lower_than_all_keys(self, smallts_withdefault):
        ts = smallts_withdefault.sample(freq=HALFHOUR, start=CURRENT - ONEHOUR)
        default = smallts_withdefault.default
        assert ts[CURRENT - ONEHOUR] == default
        assert ts[CURRENT - HALFHOUR] == default
        assert ts[CURRENT] == 0
        assert CURRENT + HALFHOUR in ts.keys()

    @pytest.mark.parametrize(
        'freq', [ONEMIN, 10 * ONEMIN,
                 timedelta(seconds=10), ONEHOUR])
    def test_sample_on_freq_with_end_and_start_in_bound(self, otherts, freq):
        start = CURRENT + 2 * ONEHOUR + HALFHOUR
        end = CURRENT + 3 * ONEHOUR + HALFHOUR
        ts = otherts.sample(freq=freq, start=start, end=end)
        assert list(ts.keys()) == [
            start + i * freq for i in range(int((end - start) / freq))
        ]



import logging
import operator
from copy import deepcopy
from datetime import timedelta

from sortedcontainers import SortedDict

from .pandas_ext import timestamp_converter

logger = logging.getLogger(__name__)


def operation_factory(operation):
    def fn_operation(self, other):
        return self._operate(other, getattr(operator, operation))

    return fn_operation


class TimeSeries(SortedDict):
    _default_interpolate = "previous"

    @property
    def lower_bound(self):
        return self.keys()[0]

    @property
    def upper_bound(self):
        return self.keys()[-1]

    def __init__(self, *args, **kwargs):
        self.default = kwargs.pop('default', None)
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            return self.set_interval(key.start, key.stop, value)
        key = timestamp_converter(key)
        super().__setitem__(key, value)

    def __getitem__(self, key):
        """Get the value of the time series, even in-between measured values by interpolation.
        Args:
            key (datetime): datetime index
            interpolate (str): interpolate operator among ["previous", "linear"]
        """

        interpolate = self._default_interpolate

        if hasattr(key, '__len__'):
            if len(key) == 2:
                key, interpolate = key
            elif len(key) > 2:
                raise KeyError

        if isinstance(key, slice):
            return self.slice(key.start, key.stop, interpolate)

        basemsg = "Getting {} but default attribute is not set".format(key)
        if self.empty:
            if self.default:
                return self.default
            else:
                raise KeyError("{} and timeseries is empty".format(basemsg))

        if key < self.lower_bound:
            if self.default:
                return self.default
            else:
                msg = "{}, can't deduce value before the oldest measurement"
                raise KeyError(msg.format(basemsg))

        # If the key is already defined:
        if key in self.keys():
            return super().__getitem__(key)

        if interpolate.lower() == "previous":
            fn = self._get_previous
        elif interpolate.lower() == "linear":
            fn = self._get_linear_interpolate
        else:
            raise ValueError("'{}' interpolation unknown.".format(interpolate))

        key = timestamp_converter(key)
        return fn(key)

    def _get_previous(self, time):
        # In this case, bisect_left == bisect_right
        # And idx > 0 as we already handled other cases
        previous_idx = self.bisect_left(time) - 1
        time_idx = self.keys()[previous_idx]
        return super().__getitem__(time_idx)

    def _get_linear_interpolate(self, time):
        idx = self.bisect_left(time)
        previous_time_idx = self.keys()[idx - 1]

        # out of right bound case:
        if idx == len(self):
            msg = "Can't interpolate out of right bound, returning value of right bound."
            logger.info(msg)
            return super().__getitem__(previous_time_idx)

        next_time_idx = self.keys()[idx]

        previous_value = super().__getitem__(previous_time_idx)
        next_value = super().__getitem__(next_time_idx)

        coeff = (time - previous_time_idx) / (
            next_time_idx - previous_time_idx)

        value = previous_value + coeff * (next_value - previous_value)
        return value

    def slice(self, start, end, interpolate=_default_interpolate):  # noqa A003
        start = timestamp_converter(start)
        end = timestamp_converter(end)

        newts = TimeSeries(default=self.default)

        for key, value in self.items():
            if start <= key and end > key:
                newts[key] = value

        should_add_left_closure = (start not in newts.keys()
                                   and not newts.empty
                                   and start >= self.lower_bound)
        if should_add_left_closure:
            newts[start] = self[start]

        return newts

    def set_interval(self, start, end, value):
        if not self.default:
            msg = "At the moment, you have to set a default for set_interval"
            raise NotImplementedError(msg)

        sliced_ts = self.slice(start, end)
        self[start] = value

        if self.empty:
            self[end] = self.default
            return

        end_is_key = end in self.keys()
        if sliced_ts.empty:
            if self.default and not end_is_key:
                self[end] = self.default
            return

        last_value_in_bound = sliced_ts[sliced_ts.upper_bound]
        for key in sliced_ts.keys():
            self.pop(key)

        self[start] = value  # may have been popped
        if not end_is_key:  # only assign if not already defined
            self[end] = last_value_in_bound

    def _operate(self, other, operator):
        if isinstance(other, self.__class__):
            return self._operate_on_ts(other, operator)
        else:
            return self._operate_on_one_value(other, operator)

    def _operate_on_ts(self, other, operator):
        if not isinstance(other, self.__class__):
            raise TypeError

        all_keys = set(self.keys()).union(set(other.keys()))

        default = None
        if self.default and other.default:
            default = operator(self.default, other.default)

        ts = TimeSeries(default=default)
        for key in all_keys:
            ts[key] = operator(self[key], other[key])

        return ts

    def _operate_on_one_value(self, value, operator):
        sample_value = self.values()[0]
        try:
            operator(value, sample_value)
        except Exception:
            msg = "Can't apply {} on {} with {}"
            raise TypeError(
                msg.format(operator.__name__, type(sample_value), type(value)))

        default = None
        if self.default:
            default = operator(self.default, value)

        ts = TimeSeries(default=default)
        for key in self.keys():
            ts[key] = operator(self[key], value)

        return ts

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("Can't compare TimeSeries with {}".format(
                type(other)))
        return super().__eq__(other) and self.default == other.default

    def __copy__(self):
        newone = super().__copy__()
        newone.default = self.default
        return newone

    def __deepcopy__(self, memo):
        newone = deepcopy(dict(self))
        newone = TimeSeries(newone, default=self.default)
        return newone

    __add__ = operation_factory('__add__')
    __radd__ = operation_factory('__add__')

    __sub__ = operation_factory('__sub__')
    __mul__ = operation_factory('__mul__')
    __div__ = operation_factory('__div__')
    __lt__ = operation_factory('__lt__')
    __le__ = operation_factory('__le__')
    __gt__ = operation_factory('__gt__')
    __ge__ = operation_factory('__ge__')

    def floor(self, other):
        return self._operate(other, min)

    def ceil(self, other):
        return self._operate(other, max)

    def conditional_update(self, other, condition):
        if not isinstance(other, TimeSeries):
            msg = 'other should be of type TimeSeries, got {}'
            raise TypeError(msg.format(type(other)))
        if not all([isinstance(value, bool) for value in condition.values()]):
            msg = 'The values of condition should all be boolean.'
            raise TypeError(msg)

        for key, value in other.items():
            if condition[key]:
                self[key] = value

    @property
    def empty(self):
        return len(self) == 0

    def compact(self):
        """Convert this instance to a compact version: consecutive measurement of the
        same value are discarded."""
        ts = TimeSeries(default=self.default)
        for time, value in self.items():
            should_set_it = ts.empty or (ts[time] != value)
            if should_set_it:
                ts[time] = value
        return ts

    def sample(self,
               freq,
               start=None,
               end=None,
               interpolate=_default_interpolate):
        if not isinstance(freq, timedelta):
            msg = 'Freq should be of instance timedelta, got {}'
            raise TypeError(msg.format(type(freq)))

        if not start:
            start = self.lower_bound
        if not end:
            # Assumption last interval is [end : end + freq[
            end = self.upper_bound + freq

        ts = TimeSeries(default=self.default)
        for i in range(0, int((end - start) / freq)):
            dt = start + i * freq
            ts[dt] = self[dt, interpolate]

        return ts



import os

from pkg_resources import DistributionNotFound, get_distribution

from . import pandas_ext
from .timeseries import TimeSeries

try:
    _dist = get_distribution("ticts")

    # Normalize case for Windows systems
    dist_loc = os.path.normcase(_dist.location)
    here = os.path.normcase(__file__)

    if not here.startswith(os.path.join(dist_loc, "ticts")):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = "Version not found."
else:
    __version__ = _dist.version

if pandas_ext.PANDAS_IS_INSTALLED:
    TimeSeries.to_dataframe = pandas_ext.to_dataframe
